/*	ptunnel.h
	Copyright (c) 2004-2005, Daniel Stoedle <daniels@stud.cs.uit.no>,
	Yellow Lemon Software. All rights reserved.
	
	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright notice,
	  this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright notice,
	  this list of conditions and the following disclaimer in the documentation
	  and/or other materials provided with the distribution.

	- Neither the name of the Yellow Lemon Software nor the names of its
	  contributors may be used to endorse or promote products derived from this
	  software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef PING_TUNNEL_H
	#define PING_TUNNEL_H

//	Includes
	#include	<pcap.h>
	
//	Constants
#define	false		0
#define	true		1
#define	bool		char

enum {
	kOpt_undefined			= 0,		//	Constants for parsing options
	kOpt_set_proxy_addr,
	kOpt_set_mode,
	kOpt_set_password,
	kOpt_set_tcp_port,
	kOpt_set_tcp_dest_addr,
	kOpt_set_tcp_dest_port,
	kOpt_set_verbosity,
	kOpt_set_max_threads,
	kOpt_set_non_privileged,
	kOpt_set_pcap_device,
	
	kMode_forward			= 0,
	kMode_proxy,
	
	kMax_threads			= 10,		/*	Set this constant to the number of concurrent
									connections you wish to handle by default. */
	kNo_log					= -1,
	kLog_error				= 0,
	kLog_info,
	kLog_event,
	kLog_verbose,
	kLog_debug,
	
	kMajor_version			= 0,
	kMinor_version			= 52,
	
	kIP_packet_max_size		= 576,
	kIP_header_size			= 20,
	kIP_actual_size			= (kIP_packet_max_size - kIP_header_size) - ((kIP_packet_max_size - kIP_header_size) % 8),
	kICMP_header_size		= 8,
	
	kDefault_buf_size		= 1024,
	
	kICMP_echo_request	= 8,
	kICMP_echo_reply		= 0,
	
	kPing_window_size		= 64,	// number of packets we can have in our send/receive ring
	
	/*	These constants are used to indicate the protocol state. The protocol
		works as follows:
		- The identifier is used by both the proxy and the forwarder
		to identify the session (and thus the relevant sockets).
		- The seq-no of the ping packet is used in a sliding-window-esque
		way, and to identify the order of data.
		
		The protocol can be in any of the following states:
		kProxy_start		Causes the proxy to open a connection to the given
							host and port, associating the ID with the socket,
							before the data on the socket are transmitted.
		kProxy_data			Indicates that the packet contains data from the proxy.
							Data ordering is indicated by the seq-no, which will start
							at 0. (The proxy and forwarder maintain different seq-nos.)
		kUser_data			This packet contains user data.
		kConnection_close	Indicates that the connection is being closed.
		kProxy_ack and		Acknowledges the packet (and all packets before it) with seq_no = ack.
		kUser_ack			This is used if there are no implicit acknowledgements due to data
							being sent.
		kProxy_req_resend	Request a resend, starting from seq_no = ack.
		kUser_req_resend	Same as above, but sent by users.
		
		Acknowledgements work by the remote peer acknowledging the last
		continuous seq no it has received.
		
		Note: A proxy receiving a kProxy_data packet, or a user receiving a
		kUser_data packet, should ignore it, as it is the host operating system
		actually returning the ping. This is mostly relevant for users, and for
		proxies running in unprivileged mode.
		
	*/
	kProxy_start			= 0,
	kProto_ack,
	kProto_data,
	kProto_close,
	kUser_flag				= 1 << 30,	//	set when packet comes from a user
	kProxy_flag				= 1 << 31,	//	set when packet comes from the proxy
	kFlag_mask				= kUser_flag | kProxy_flag,
};

#define	kPing_tunnel_magic		0xD5200880
//	Resend packets after this interval (in seconds)
#define	kResend_interval		1.5

#pragma pack(push, 1)

typedef struct {
	unsigned long	magic;
	long				  dst_ip,
						    dst_port,
						    state,
						    ack,		//	only used in packets from client to proxy
						    data_len;
	uint16_t			seq_no,
						    id_no;
	char				  data[0];
} ping_tunnel_pkt_t;


typedef struct {
	unsigned char	  vers_ihl,
					        tos;
	unsigned short	pkt_len,
					        id,
					        flags_frag_offset;
	unsigned char	  ttl,
					        proto;	// 1 for ICMP
	unsigned short	checksum;
	unsigned long	  src_ip,
					        dst_ip;
	char			      data[0];
} ip_packet_t;


typedef struct {
	unsigned char	  type,	// 3
					        code;
	unsigned short	checksum,
					        identifier,
					        seq;
	char			      data[0];
} icmp_echo_packet_t;

#pragma pack(pop)

typedef struct {
	int		sock;
} pt_thread_info_t;


typedef struct {
	int		seq_no,
				length,
				remaining;
	char	data[0];
} forward_desc_t;


typedef struct {
	int				pkt_len;
	double		last_resend,
						jitter;
	int				resend_count;
	uint16_t	seq_no,
            icmp_id;
	icmp_echo_packet_t	*pkt;
} icmp_desc_t;


typedef struct proxy_desc_t {
	int				sock,
						bytes,
						should_remove;
	char			*buf;
	uint16_t	id_no,
						my_seq,
						ping_seq,
						next_remote_seq,
						pkt_type,
						remote_ack_val,
            icmp_id;
	int				recv_idx,		//	first available slot in recv ring
						recv_xfer_idx,	//	current slot in recv ring being transferred
						send_idx,		//	first available slot in send ring
						send_first_ack,	//	first packet in send ring not yet acked
						recv_wait_send,	//	number of items in recv ring awaiting send
						send_wait_ack,	//	number of items in send ring awaiting ack
						next_resend_start;
	uint32_t	state,
						type_flag,		//	Either kProxy_flag or kUser_flag
						dst_ip,
						dst_port;
	double		        	last_ack;		//	Time when last ack packet was sent.
	struct sockaddr_in	dest_addr;
	icmp_desc_t			    send_ring[kPing_window_size];
	forward_desc_t		  *recv_ring[kPing_window_size];
	struct proxy_desc_t	*next;
} proxy_desc_t;


typedef struct pqueue_elem_t {
	int						bytes;
	struct pqueue_elem_t	*next;
	char					data[0];
} pqueue_elem_t;


typedef struct {
	pqueue_elem_t	*head,
					*tail;
	int				elems;
} pqueue_t;


typedef struct {
	pcap_t				*pcap_desc;
	struct bpf_program	fp;		//	Compiled filter program
	uint32_t			netp,
						netmask;
	char				*pcap_err_buf,
						*pcap_data_buf;
	pqueue_t			pkt_q;
} pcap_info_t;


//	Prototypes
	void		usage(char *exec_name);
	void*		pt_proxy(void *args);
	void		pcap_packet_handler(u_char *refcon, const struct pcap_pkthdr *hdr, const u_char* pkt);
	void		handle_packet(char *buf, int bytes, int is_pcap, struct sockaddr_in *addr);
	
	proxy_desc_t*	create_and_insert_proxy_desc(uint16_t id_no, uint16_t icmp_id, int sock, struct sockaddr_in *addr, uint32_t dst_ip, uint32_t dst_port, uint32_t init_state, uint32_t type);
	void		remove_proxy_desc(proxy_desc_t *cur, proxy_desc_t *prev);
	
	void		pt_forwarder(void);
	
	int			queue_packet(int icmp_sock, uint8_t type, char *buf, int num_bytes, uint16_t id_no, uint16_t icmp_id, uint16_t *seq, icmp_desc_t ring[], int *insert_idx, int *await_send, uint32_t ip, uint32_t port, uint32_t state, struct sockaddr_in *dest_addr, uint16_t next_expected_seq, int *first_ack, uint16_t *ping_seq);
	void		send_packets(forward_desc_t *ring[], int *xfer_idx, int *await_send, int *sock);
	void		handle_data(icmp_echo_packet_t *pkt, int total_len, forward_desc_t *ring[], int *await_send, int *insert_idx, uint16_t *next_expected_seq);
	void		handle_ack(uint16_t seq_no, icmp_desc_t ring[], int *packets_awaiting_ack, int one_ack_only, int insert_idx, int *first_ack, uint16_t *remote_ack, int is_pcap);
	forward_desc_t*	create_fwd_desc(uint16_t seq_no, uint32_t data_len, char *data);
//	ip_packet_t**		make_data_packets(char *data, int data_len, unsigned long src_ip, unsigned long dst_ip, unsigned short src_port, unsigned short dst_port, int *num_frags);
	void			init_ip_packet(ip_packet_t *packet, unsigned short id, unsigned short frag_offset, unsigned short pkt_len, unsigned char ttl, unsigned long src_ip, unsigned long dst_ip, bool is_last_frag, bool dont_frag);
	unsigned short	calc_ip_checksum(ip_packet_t *pkt);
	unsigned short	calc_icmp_checksum(unsigned short *data, int bytes);
	
	char*	f_inet_ntoa(long ip);
	void	pt_log(int level, char *fmt, ...);
	double	timeval_to_double(void);
	void FillErrorText(DWORD dwLastError, char * buf);
#endif
